"""A virtual tile that ensapsulates resuable behavior."""

from iotile.core.exceptions import ArgumentError
from iotile.core.dev import ComponentRegistry
from .base_runnable import BaseRunnable
from .common_types import tile_rpc, RPCDispatcher


class VirtualTile(BaseRunnable, RPCDispatcher):
    """A virtual tile.

    Tiles have their own RPC based API and can run background
    threads to do periodic work.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        name (str): The 6 character name that should be returned when this
            tile is asked for its status to allow matching it with a proxy
            object.
        device (TileBasedVirtualDevice) : optional, device on which this tile is running
    """

    def __init__(self, address, name, device=None):
        super(VirtualTile, self).__init__()

        self.address = address
        self.name = self._check_convert_name(name)

    @classmethod
    def _check_convert_name(cls, name):
        if not isinstance(name, bytes):
            name = name.encode('utf-8')
        if len(name) < 6:
            name += b' '*(6 - len(name))
        elif len(name) > 6:
            raise ArgumentError("Virtual tile name is too long, it must be 6 or fewer characters")

        return name

    def start(self, channel=None):
        """Start any background workers on this tile."""
        self.start_workers()

    def stop(self):
        """Stop any background workers on this tile."""
        self.stop_workers()

    def signal_stop(self):
        """Asynchronously signal that all workers should stop."""
        self.stop_workers_async()

    def wait_stopped(self):
        """Wait for all workers to stop."""
        self.wait_workers_stopped()

    @classmethod
    def FindByName(cls, name):
        """Find an installed VirtualTile by name.

        This function searches for installed virtual tiles
        using the pkg_resources entry_point `iotile.virtual_tile`.

        If name is a path ending in .py, it is assumed to point to
        a module on disk and loaded directly rather than using
        pkg_resources.

        Args:
            name (str): The name of the tile to search
                for.

        Returns:
            VirtualTile class: A virtual tile subclass that can be
                instantiated to create a virtual tile.
        """

        if name.endswith('.py'):
            return cls.LoadFromFile(name)

        reg = ComponentRegistry()
        for _name, tile in reg.load_extensions('iotile.virtual_tile', name_filter=name, class_filter=VirtualTile):
            return tile

        raise ArgumentError("VirtualTile could not be found by name", name=name)

    @classmethod
    def LoadFromFile(cls, script_path):
        """Import a virtual tile from a file rather than an installed module

        script_path must point to a python file ending in .py that contains exactly one
        VirtualTile class definition.  That class is loaded and executed as if it
        were installed.

        To facilitate development, if there is a proxy object defined in the same
        file, it is also added to the HardwareManager proxy registry so that it
        can be found and used with the device.

        Args:
            script_path (string): The path to the script to load

        Returns:
            VirtualTile: A subclass of VirtualTile that was loaded from script_path
        """

        _name, dev = ComponentRegistry().load_extension(script_path, class_filter=VirtualTile, unique=True)
        return dev

    @tile_rpc(0x0004, "", "H6sBBBB")
    def tile_status(self):
        """Required status RPC that allows matching a proxy object with a tile."""

        status = (1 << 1) | (1 << 0)  # Configured and running, not currently used but required for compat with physical tiles
        return [0xFFFF, self.name, 1, 0, 0, status]



# Annotated Registry
# A wrapper to make the IOTile component registry accessible in the iotile
# tool.  Since the registry is used internally in the type system it cannot
# itself make use of typedargs annotations
from iotile.core.utilities.typedargs import annotated, param, return_type, context
from iotile.core.exceptions import ExternalError
from .registry import ComponentRegistry

_name_ = "Developer"


@annotated
def registry():
    return AnnotatedRegistry()

@context()
class AnnotatedRegistry(object):
    """
    AnnotatedRegistry

    A mapping of all of the installed components on this system that can
    be used as build dependencies and where they are located.  Also used
    to manage iotile plugins.
    """

    def __init__(self):
        self.reg = ComponentRegistry()

    @param("component", "path", "exists", desc="folder containing component to register")
    def add_component(self, component):
        """
        Register a component with ComponentRegistry.

        Component must be a buildable object with a module_settings.json file that
        describes its name and the domain that it is part of.
        """

        self.reg.add_component(component)

    @return_type("map(string, string)")
    def list_plugins(self):
        """
        List all of the plugins that have been registerd for the iotile program on this computer
        """

        return self.reg.list_plugins()

    def find_component(self, key, domain=""):
        return self.reg.find_component(key, domain)

    @param("key", "string", desc="iotile component to find")
    def remove_component(self, key):
        """
        Remove component from registry
        """

        return self.reg.remove_component(key)

    @annotated
    def clear_components(self):
        """
        Clear all of the registered components
        """

        self.reg.clear_components()

    @return_type("list(string)")
    def list_components(self):
        """
        List all of the registered component names
        """

        return self.reg.list_components()

    @return_type("list(string)")
    def list_config(self):
        """
        List all of the registered component names
        """

        return self.reg.list_config()

    @param("key", "string")
    @param("value", "string")
    def set(self, key, value):
        self.reg.set_config(key, value)

    @param("key", "string")
    @return_type("string")
    def get(self, key):
        return self.reg.get_config(key)

    @param("key", "string")
    @return_type("string")
    def clear(self, key):
        self.reg.clear_config(key)

    @annotated
    def freeze(self):
        """Freeze the current list of extensions to a single file.

        This speeds up the extension loading process but does not check for
        new extensions each time the program is run. You should only use this
        method in situations where you know the extension list is static
        and you need the speedup benefits.

        You can undo this by calling unfreeze().
        """

        self.reg.freeze_extensions()

    @annotated
    def unfreeze(self):
        """Remove any frozen extension list."""

        try:
            self.reg.unfreeze_extensions()
        except ExternalError:
            pass  # This means there was no frozen list of extensions



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import itertools
from collections import namedtuple
from string import Template
import json
import os.path
import sys
from future.utils import viewitems, itervalues

if sys.version_info >= (3, 0):
    basestring = str

from iotile.core.exceptions import DataError, ExternalError
from .semver import SemanticVersion, SemanticVersionRange

ReleaseStep = namedtuple('ReleaseStep', ['provider', 'args'])
ReleaseInfo = namedtuple('ReleaseInfo', ['release_date', 'dependency_versions'])
TileInfo = namedtuple('TileInfo', ['module_name', 'settings', 'architectures', 'targets', 'release_data'])

_ProductDeclaration = namedtuple('_ProductDeclaration', ['dev_path', 'release_path', 'process_func'])
_ReleaseOnlyProduct = _ProductDeclaration(r"${release}/${product}", r"${release}/${product}", None)
_DevOnlyProduct = _ProductDeclaration(r"${module}/${product}", r"${module}/${product}", None)


class IOTile(object):
    """Object representing an IOTile component that implements or extends CoreTools.

    IOTile components are "projects" in the sense of traditional IDEs,
    "packages" in nodejs or "distributions" in python.  They are single
    directories that contain related code and a metadata file that describes
    what the project is and how to use it.

    Since IOTile components describe building blocks for IOT devies, they do not
    contain code in a single language but could describe a variety of different
    kinds of artifacts such as:

    - firmware images
    - python extensions for CoreTools
    - build automation steps used in hardware manufacturing
    - emulators for specific IOTile devices

    The unifying concepts that extend to all of the above finds of components are
    that they product "artifacts" through a build process, the entire component
    has a single version and it can have dependencies on other components.

    There is a single file inside the root of the component directory,
    `module_settings.json` that contains all of this metadata.

    The `IOTile` object is constructed from a `module_settings.json` file and
    provides helper routines to parse and search its contents.

    Args:
        folder (str): The path to a folder containing a `module_settings.json`
            file that will be loaded into this IOTile object.
    """

    V1_FORMAT = "v1"
    V2_FORMAT = "v2"

    PYTHON_PRODUCTS = frozenset([
        "build_step",
        "app_module",
        "proxy_module",
        "type_package",
        "proxy_plugin",
        "virtual_tile",
        "virtual_device"
    ])
    """The canonical list of all product types that contain python code.

    This is used to know if this IOTile component will contain a support
    wheel, which happens if it produces any products of these types.
    """

    LIST_PRODUCTS = frozenset([
        "include_directories",
        "tilebus_definitions"
    ])
    """The canonical list of products that are stored as a list.

    Most products are stored in module_settings.json in the products map where
    the key is the path to the product and the value is the type of product.
    Some products, those in this property, are stored where the key is the
    type of product and all of the specific products are stored in a single
    list under that key.
    """

    PATH_PRODUCTS = {
        "include_directories": _ProductDeclaration(r"${release}/include/${product}", r"${release}/include/${product}", None),
        "tilebus_definitions": _ProductDeclaration(r"${module}/firmware/src/${raw_product}", r"${release}/tilebus/${product}", os.path.basename),
        "linker_script": _ProductDeclaration(r"${release}/linker/${product}", r"${release}/linker/${product}", None),
        "type_package": _DevOnlyProduct,
        "build_step": _DevOnlyProduct,
        "app_module": _DevOnlyProduct,
        "proxy_module": _DevOnlyProduct,
        "proxy_plugin": _DevOnlyProduct,
        "virtual_tile": _DevOnlyProduct,
        "virtual_device": _DevOnlyProduct,
        "firmware_image": _ReleaseOnlyProduct
    }
    """Declarations for products that require special path processing."""

    def __init__(self, folder):
        self.folder = folder
        self.filter_prods = False

        modfile = os.path.join(self.folder, 'module_settings.json')

        try:
            with open(modfile, "r") as infile:
                settings = json.load(infile)
        except IOError:
            raise ExternalError("Could not load module_settings.json file, make sure this directory is an IOTile component", path=self.folder)

        file_format = settings.get('file_format', IOTile.V1_FORMAT)
        if file_format == IOTile.V1_FORMAT:
            info = self._find_v1_settings(settings)
        elif file_format == IOTile.V2_FORMAT:
            info = self._find_v2_settings(settings)
        else:
            raise DataError("Unknown file format in module_settings.json", format=file_format, path=modfile)

        self._load_settings(info)

    def _find_v1_settings(self, settings):
        """Parse a v1 module_settings.json file.

        V1 is the older file format that requires a modules dictionary with a
        module_name and modules key that could in theory hold information on
        multiple modules in a single directory.
        """

        if 'module_name' in settings:
            modname = settings['module_name']
        if 'modules' not in settings or len(settings['modules']) == 0:
            raise DataError("No modules defined in module_settings.json file")
        elif len(settings['modules']) > 1:
            raise DataError("Multiple modules defined in module_settings.json file", modules=[x for x in settings['modules']])
        else:
            modname = list(settings['modules'])[0]

        if modname not in settings['modules']:
            raise DataError("Module name does not correspond with an entry in the modules directory", name=modname, modules=[x for x in settings['modules']])

        release_info = self._load_release_info(settings)
        modsettings = settings['modules'][modname]
        architectures = settings.get('architectures', {})

        target_defs = settings.get('module_targets', {})
        targets = target_defs.get(modname, [])

        return TileInfo(modname, modsettings, architectures, targets, release_info)

    @classmethod
    def _load_release_info(cls, settings):
        if settings.get('release', False) is False:
            return None

        if 'release_date' not in settings:
            raise DataError("Release mode IOTile component did not include a release date")

        import dateutil.parser
        release_date = dateutil.parser.parse(settings['release_date'])
        dependency_versions = {x: SemanticVersion.FromString(y) for x, y in viewitems(settings.get('dependency_versions', {}))}

        return ReleaseInfo(release_date, dependency_versions)

    def _find_v2_settings(self, settings):
        archs = settings.get('architectures', {})
        mod_name = settings.get('module_name')
        release_info = self._load_release_info(settings)

        targets = settings.get('targets', [])

        return TileInfo(mod_name, settings, archs, targets, release_info)

    def _load_settings(self, info):
        """Load settings for a module."""

        modname, modsettings, architectures, targets, release_info = info
        self.settings = modsettings

        #Name is converted to all lowercase to canonicalize it
        prepend = ''
        if 'domain' in modsettings:
            prepend = modsettings['domain'].lower() + '/'

        key = prepend + modname.lower()

        #Copy over some key properties that we want easy access to
        self.name = key
        self.unique_id = key.replace('/', '_')
        self.short_name = modname
        self.targets = targets

        self.full_name = "Undefined"
        if "full_name" in self.settings:
            self.full_name = self.settings['full_name']

        #FIXME: make sure this is a list
        self.authors = []
        if "authors" in self.settings:
            self.authors = self.settings['authors']

        self.version = "0.0.0"
        if "version" in self.settings:
            self.version = self.settings['version']

        self.parsed_version = SemanticVersion.FromString(self.version)

        #Load all of the build products that can be created by this IOTile
        self.products = modsettings.get('products', {})

        #Load in the release information telling us how to release this component
        release_steps = modsettings.get('release_steps', [])
        self.release_steps = []
        self.can_release = False

        for step in release_steps:
            if 'provider' not in step:
                raise DataError("Invalid release step that did not have a provider key", step=step)

            parsed_step = ReleaseStep(provider=step['provider'], args=step.get('args', {}))
            self.release_steps.append(parsed_step)

        if len(self.release_steps) > 0:
            self.can_release = True

        self.dependency_versions = {}

        #If this is a release IOTile component, check for release information
        if release_info is not None:
            self.release = True
            self.release_date = release_info.release_date
            self.output_folder = self.folder
            self.dependency_versions = release_info.dependency_versions
        else:
            self.release = False
            self.output_folder = os.path.join(self.folder, 'build', 'output')

            #If this tile is a development tile and it has been built at least one, add in a release date
            #from the last time it was built
            if os.path.exists(os.path.join(self.output_folder, 'module_settings.json')):
                release_settings = os.path.join(self.output_folder, 'module_settings.json')

                with open(release_settings, 'r') as infile:
                    release_dict = json.load(infile)

                import dateutil.parser
                self.release_date = dateutil.parser.parse(release_dict['release_date'])
            else:
                self.release_date = None

        #Find all of the things that this module could possibly depend on
        #Dependencies include those defined in the module itself as well as
        #those defined in architectures that are present in the module_settings.json
        #file.
        self.dependencies = []

        archs_with_deps = [viewitems(y['depends']) for _x, y in viewitems(architectures) if 'depends' in y]
        if 'depends' in self.settings:
            if not isinstance(self.settings['depends'], dict):
                raise DataError("module must have a depends key that is a dictionary", found=str(self.settings['depends']))

            archs_with_deps.append(viewitems(self.settings['depends']))

        #Find all python package needed
        self.support_wheel_depends = []

        if 'python_depends' in self.settings:
            if not isinstance(self.settings['python_depends'], list):
                raise DataError("module must have a python_depends key that is a list of strings",
                                found=str(self.settings['python_depends']))

            for python_depend in self.settings['python_depends']:
                if not isinstance(python_depend, basestring):
                    raise DataError("module must have a python_depends key that is a list of strings",
                                    found=str(self.settings['python_depends']))

                self.support_wheel_depends.append(python_depend)

        #Also search through overlays to architectures that are defined in this module_settings.json file
        #and see if those overlays contain dependencies.
        for overlay_arch in itervalues(self.settings.get('overlays', {})):
            if 'depends' in overlay_arch:
                archs_with_deps.append(viewitems(overlay_arch['depends']))

        found_deps = set()
        for dep, _ in itertools.chain(*archs_with_deps):
            name, _, version = dep.partition(',')
            unique_id = name.lower().replace('/', '_')

            version = version.strip()
            if version == '':
                version = "*"

            ver_range = SemanticVersionRange.FromString(version)

            depdict = {
                'name': name,
                'unique_id': unique_id,
                'required_version': ver_range,
                'required_version_string': version
            }

            if name not in found_deps:
                self.dependencies.append(depdict)

            found_deps.add(name)

        # Store any architectures that we find in this json file for future reference
        self.architectures = architectures

        # Setup our support package information
        self.support_distribution = "iotile_support_{0}_{1}".format(self.short_name, self.parsed_version.major)

        if 'python_universal' in self.settings:
            py_version = "py2.py3"
        elif sys.version_info[0] >= 3:
            py_version = "py3"
        else:
            py_version = "py2"

        self.support_wheel = "{0}-{1}-{2}-none-any.whl".format(self.support_distribution,
                                                               self.parsed_version.pep440_string(),
                                                               py_version)
        self.has_wheel = self._check_has_wheel()

    def _check_has_wheel(self):
        for prod in self.PYTHON_PRODUCTS:
            if len(self.find_products(prod)) > 0:
                return True

        return False

    @classmethod
    def _ensure_product_string(cls, product):
        """Ensure that all product locations are strings.

        Older components specify paths as lists of path components.  Join
        those paths into a normal path string.
        """

        if isinstance(product, basestring):
            return product

        if isinstance(product, list):
            return os.path.join(*product)

        raise DataError("Unknown object (not str or list) specified as a component product", product=product)

    def _process_product_path(self, product, declaration):
        processed = product
        if declaration.process_func is not None:
            processed = declaration.process_func(product)

        if self.release:
            base = Template(declaration.release_path)
        else:
            base = Template(declaration.dev_path)

        path_string = base.substitute(module=self.folder, release=self.output_folder, raw_product=product, product=processed)
        return os.path.normpath(path_string)

    def find_products(self, product_type):
        """Search for products of a given type.

        Search through the products declared by this IOTile component and
        return only those matching the given type.  If the product is decribed
        by the path to a file, a complete normalized path will be returned.
        The path could be different depending on whether this IOTile component
        is in development or release mode.

        The behavior of this function when filter_products has been called is
        slightly different based on whether product_type is in LIST_PRODUCTS
        or not.  If product type is in LIST_PRODUCTS, then all matching
        products are returned if product_type itself was passed.  So to get
        all tilebus_definitions you would call
        ``filter_products('tilebus_definitions')``

        By constast, other products are filtered product-by-product.  So there
        is no way to filter and get **all libraries**.  Instead you pass the
        specific product names of the libraries that you want to
        ``filter_products`` and those specific libraries are returned.
        Passing the literal string ``library`` to ``filter_products`` will not
        return only the libraries, it will return nothing since no library is
        named ``library``.

        Args:
            product_type (str): The type of product that we wish to return.

        Returns:
            list of str: The list of all products of the given type.

            If no such products are found, an empty list will be returned.
            If filter_products() has been called and the filter does not include
            this product type, an empty list will be returned.
        """

        if self.filter_prods and product_type in self.LIST_PRODUCTS and product_type not in self.desired_prods:
            return []

        if product_type in self.LIST_PRODUCTS:
            found_products = self.products.get(product_type, [])
        else:
            found_products = [x[0] for x in viewitems(self.products)
                              if x[1] == product_type and (not self.filter_prods or x[0] in self.desired_prods)]

        found_products = [self._ensure_product_string(x) for x in found_products]

        declaration = self.PATH_PRODUCTS.get(product_type)
        if declaration is not None:
            found_products = [self._process_product_path(x, declaration) for x in found_products]

        return found_products

    def library_directories(self):
        """Return a list of directories containing any static libraries built by this IOTile."""

        libs = self.find_products('library')

        if len(libs) > 0:
            return [os.path.join(self.output_folder)]

        return []

    def filter_products(self, desired_prods):
        """When asked for a product, filter only those on this list."""

        self.filter_prods = True
        self.desired_prods = set(desired_prods)

    def path(self):
        """The path to this component."""
        return self.folder



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import os.path
import sys
import logging
import imp
import inspect
import json
from types import ModuleType
from future.utils import itervalues

if sys.version_info >= (3, 0):
    basestring = str

import entrypoints

from iotile.core.utilities.kvstore_sqlite import SQLiteKVStore
from iotile.core.utilities.kvstore_json import JSONKVStore
from iotile.core.utilities.kvstore_mem import InMemoryKVStore
from iotile.core.exceptions import ArgumentError, ExternalError
from iotile.core.utilities.paths import settings_directory
from .iotileobj import IOTile

MISSING = object()

class ComponentRegistry(object):
    """
    ComponentRegistry

    A mapping of all of the installed components on this system that can
    be used as build dependencies and where they are located.  Also used
    to manage iotile plugins.
    """

    BackingType = SQLiteKVStore
    BackingFileName = 'component_registry.db'

    _registered_extensions = {}
    _component_overlays = {}
    _frozen_extensions = None

    def __init__(self):
        self._kvstore = None
        self._plugins = None
        self._logger = logging.getLogger(__name__)

    @property
    def frozen(self):
        """Return whether we have a cached list of all installed entry_points."""

        frozen_path = os.path.join(_registry_folder(), 'frozen_extensions.json')
        return os.path.isfile(frozen_path)

    @property
    def kvstore(self):
        """Lazily load the underlying key-value store backing this registry."""

        if self._kvstore is None:
            self._kvstore = self.BackingType(self.BackingFileName, respect_venv=True)

        return self._kvstore

    @property
    def plugins(self):
        """Lazily load iotile plugins only on demand.

        This is a slow operation on computers with a slow FS and is rarely
        accessed information, so only compute it when it is actually asked
        for.
        """

        if self._plugins is None:
            self._plugins = {}

            for _, plugin in self.load_extensions('iotile.plugin'):
                links = plugin()
                for name, value in links:
                    self._plugins[name] = value

        return self._plugins

    def load_extensions(self, group, name_filter=None, comp_filter=None, class_filter=None, product_name=None, unique=False):
        """Dynamically load and return extension objects of a given type.

        This is the centralized way for all parts of CoreTools to allow plugin
        behavior.  Whenever a plugin is needed, this method should be called
        to load it.  Examples of plugins are proxy modules, emulated tiles,
        iotile-build autobuilders, etc.

        Each kind of plugin will typically be a subclass of a certain base class
        and can be provided one of three ways:

        1. It can be registered as an entry point in a pip installed package.
           The entry point group must map the group passed to load_extensions.
        2. It can be listed as a product of an IOTile component that is stored
           in this ComponentRegistry.  The relevant python file inside the
           component will be imported dynamically as needed.
        3. It can be programmatically registered by calling ``register_extension()``
           on this class with a string name and an object.  This is equivalent to
           exposing that same object as an entry point with the given name.

        There is special behavior of this function if class_filter is passed
        and the object returned by one of the above three methods is a python
        module. The module will be search for object definitions that match
        the defined class.

        The order of the returned objects list is only partially defined.
        Locally installed components are searched before pip installed
        packages with entry points.  The order of results within each group is
        not specified.

        Args:
            group (str): The extension type that you wish to enumerate.  This
                will be used as the entry_point group for searching pip
                installed packages.
            name_filter (str): Only return objects with the given name
            comp_filter (str): When searching through installed components
                (not entry_points), only search through components with the
                given name.
            class_filter (type or tuple of types): An object that will be passed to
                instanceof() to verify that all extension objects have the correct
                types.  If not passed, no checking will be done.
            product_name (str): If this extension can be provided by a registered
                local component, the name of the product that should be loaded.
            unique (bool): If True (default is False), there must be exactly one object
                found inside this extension that matches all of the other criteria.

        Returns:
            list of (str, object): A list of the found and loaded extension objects.

            The string returned with each extension is the name of the
            entry_point or the base name of the file in the component or the
            value provided by the call to register_extension depending on how
            the extension was found.

            If unique is True, then the list only contains a single entry and that
            entry will be directly returned.
        """

        found_extensions = []

        if product_name is not None:
            for comp in self.iter_components():
                if comp_filter is not None and comp.name != comp_filter:
                    continue

                products = comp.find_products(product_name)
                for product in products:
                    try:
                        entries = self.load_extension(product, name_filter=name_filter, class_filter=class_filter,
                                                      component=comp)
                        if len(entries) == 0 and name_filter is None:  # Don't warn if we're filtering by name since most extensions won't match
                            self._logger.warn("Found no valid extensions in product %s of component %s", product, comp.path)
                            continue

                        found_extensions.extend(entries)
                    except:  #pylint:disable=bare-except;We don't want a broken extension to take down the whole system
                        self._logger.exception("Unable to load extension %s from local component %s at path %s", product_name, comp, product)

        for entry in self._iter_entrypoint_group(group):
            name = entry.name

            if name_filter is not None and name != name_filter:
                continue

            ext = entry.load()

            found_extensions.extend((name, x) for x in self._filter_subclasses(ext, class_filter))

        for (name, ext) in self._registered_extensions.get(group, []):
            if name_filter is not None and name != name_filter:
                continue

            found_extensions.extend((name, x) for x in self._filter_subclasses(ext, class_filter))

        found_extensions = [(name, x) for name, x in found_extensions if self._filter_nonextensions(x)]

        if unique is True:
            if len(found_extensions) > 1:
                raise ArgumentError("Extension %s should have had exactly one instance of class %s, found %d" % (group, class_filter.__name__, len(found_extensions)), classes=found_extensions)
            elif len(found_extensions) == 0:
                raise ArgumentError("Extension %s had no instances of class %s" % (group, class_filter.__name__))

            return found_extensions[0]

        return found_extensions

    def register_extension(self, group, name, extension):
        """Register an extension.

        Args:
            group (str): The type of the extension
            name (str): A name for the extension
            extension (str or class): If this is a string, then it will be
                interpreted as a path to import and load.  Otherwise it
                will be treated as the extension object itself.
        """

        if isinstance(extension, basestring):
            name, extension = self.load_extension(extension)[0]

        if group not in self._registered_extensions:
            self._registered_extensions[group] = []

        self._registered_extensions[group].append((name, extension))

    def clear_extensions(self, group=None):
        """Clear all previously registered extensions."""

        if group is None:
            self._registered_extensions = {}
            return

        if group in self._registered_extensions:
            self._registered_extensions[group] = []

    def freeze_extensions(self):
        """Freeze the set of extensions into a single file.

        Freezing extensions can speed up the extension loading process on
        machines with slow file systems since it requires only a single file
        to store all of the extensions.

        Calling this method will save a file into the current virtual
        environment that stores a list of all currently found extensions that
        have been installed as entry_points.  Future calls to
        `load_extensions` will only search the one single file containing
        frozen extensions rather than enumerating all installed distributions.
        """

        output_path = os.path.join(_registry_folder(), 'frozen_extensions.json')

        with open(output_path, "w") as outfile:
            json.dump(self._dump_extensions(), outfile)

    def unfreeze_extensions(self):
        """Remove a previously frozen list of extensions."""

        output_path = os.path.join(_registry_folder(), 'frozen_extensions.json')
        if not os.path.isfile(output_path):
            raise ExternalError("There is no frozen extension list")

        os.remove(output_path)
        ComponentRegistry._frozen_extensions = None

    def load_extension(self, path, name_filter=None, class_filter=None, unique=False, component=None):
        """Load a single python module extension.

        This function is similar to using the imp module directly to load a
        module and potentially inspecting the objects it declares to filter
        them by class.

        Args:
            path (str): The path to the python file to load
            name_filter (str): If passed, the basename of the module must match
                name or nothing is returned.
            class_filter (type): If passed, only instance of this class are returned.
            unique (bool): If True (default is False), there must be exactly one object
                found inside this extension that matches all of the other criteria.
            component (IOTile): The component that this extension comes from if it is
                loaded from an installed component.  This is used to properly import
                the extension as a submodule of the component's support package.

        Returns:
            list of (name, type): A list of the objects found at the extension path.

            If unique is True, then the list only contains a single entry and that
            entry will be directly returned.
        """

        import_name = None
        if component is not None:
            import_name = _ensure_package_loaded(path, component)

        name, ext = _try_load_module(path, import_name=import_name)

        if name_filter is not None and name != name_filter:
            return []

        found = [(name, x) for x in self._filter_subclasses(ext, class_filter)]
        found = [(name, x) for name, x in found if self._filter_nonextensions(x)]

        if not unique:
            return found

        if len(found) > 1:
            raise ArgumentError("Extension %s should have had exactly one instance of class %s, found %d" % (path, class_filter.__name__, len(found)), classes=found)
        elif len(found) == 0:
            raise ArgumentError("Extension %s had no instances of class %s" % (path, class_filter.__name__))

        return found[0]

    def _load_frozen_extensions(self):
        self._logger.critical("Loading frozen extensions from file, new extensions will not be found")

        frozen_path = os.path.join(_registry_folder(), 'frozen_extensions.json')
        with open(frozen_path, "r") as infile:
            extensions = json.load(infile)

        ComponentRegistry._frozen_extensions = {}
        for group in extensions:
            ComponentRegistry._frozen_extensions[group] = []

            for ext_info in extensions.get(group, []):
                name = ext_info['name']
                obj_path = ext_info['object']
                distro_info = ext_info['distribution']

                distro = None
                if distro_info is not None:
                    distro = entrypoints.Distribution(*distro_info)

                entry = entrypoints.EntryPoint.from_string(obj_path, name, distro=distro)
                ComponentRegistry._frozen_extensions[group].append(entry)

    def _iter_entrypoint_group(self, group):
        if not self.frozen:
            return entrypoints.get_group_all(group)

        if self._frozen_extensions is None:
            self._load_frozen_extensions()

        return self._frozen_extensions.get(group, [])

    @classmethod
    def _filter_nonextensions(cls, obj):
        """Remove all classes marked as not extensions.

        This allows us to have a deeper hierarchy of classes than just
        one base class that is filtered by _filter_subclasses.  Any
        class can define a class propery named:

        __NO_EXTENSION__ = True

        That class will never be returned as an extension.  This is useful
        for masking out base classes for extensions that are declared in
        CoreTools and would be present in module imports but should not
        create a second entry point.
        """

        # Not all objects have __dict__ attributes.  For example, tuples don't.
        # and tuples are used in iotile.build for some entry points.
        if hasattr(obj, '__dict__') and obj.__dict__.get('__NO_EXTENSION__', False) is True:
            return False

        return True

    @classmethod
    def _dump_extensions(cls, prefix="iotile."):
        extensions = {}

        for config, distro in entrypoints.iter_files_distros():
            if distro is None:
                distro_info = None
            else:
                distro_info = (distro.name, distro.version)

            for group in config:
                if prefix is not None and not group.startswith(prefix):
                    continue

                if group not in extensions:
                    extensions[group] = []

                for name, epstr in config[group].items():
                    extensions[group].append(dict(name=name, object=epstr, distribution=distro_info))

        return extensions

    def _filter_subclasses(self, obj, class_filter):
        if class_filter is None:
            return [obj]

        if isinstance(obj, ModuleType):
            return [x for x in itervalues(obj.__dict__) if inspect.isclass(x) and issubclass(x, class_filter) and x != class_filter]

        if inspect.isclass(obj) and issubclass(obj, class_filter):
            return [obj]

        self._logger.warn("Found no valid extension objects in %s, sought subclasses of %s", obj, class_filter)
        return []

    @classmethod
    def SetBackingStore(cls, backing):
        """Set the global backing type used by the ComponentRegistry from this point forward

        This function must be called before any operations that use the registry are initiated
        otherwise they will work from different registries that will likely contain different data
        """

        if backing not in ['json', 'sqlite', 'memory']:
            raise ArgumentError("Unknown backing store type that is not json or sqlite", backing=backing)

        if backing == 'json':
            cls.BackingType = JSONKVStore
            cls.BackingFileName = 'component_registry.json'
        elif backing == 'memory':
            cls.BackingType = InMemoryKVStore
            cls.BackingFileName = None
        else:
            cls.BackingType = SQLiteKVStore
            cls.BackingFileName = 'component_registry.db'

    def add_component(self, component, temporary=False):
        """Register a component with ComponentRegistry.

        Component must be a buildable object with a module_settings.json file
        that describes its name and the domain that it is part of.  By
        default, this component is saved in the permanent registry associated
        with this environment and will remain registered for future CoreTools
        invocations.

        If you only want this component to be temporarily registered during
        this program's session, you can pass temporary=True and the component
        will be stored in RAM only, not persisted to the underlying key-value
        store.

        Args:
            component (str): The path to a component that should be registered.
            temporary (bool): Optional flag to only temporarily register the
                component for the duration of this program invocation.
        """

        tile = IOTile(component)
        value = os.path.normpath(os.path.abspath(component))

        if temporary is True:
            self._component_overlays[tile.name] = value
        else:
            self.kvstore.set(tile.name, value)

    def get_component(self, component):
        if component in self._component_overlays:
            return IOTile(self._component_overlays[component])

        try:
            comp_path = self.kvstore.get(component)
        except KeyError:
            raise ArgumentError("Could not find component by name", component=component)

        return IOTile(comp_path)

    def list_plugins(self):
        """
        List all of the plugins that have been registerd for the iotile program on this computer
        """

        vals = self.plugins.items()

        return {x: y for x, y in vals}

    def find_component(self, key, domain=""):
        try:
            if len(domain) != 0:
                key = domain.lower() + '/' + key.lower()

            return IOTile(self.kvstore.get(key))
        except KeyError:
            raise ArgumentError("Unknown component name", name=key)

    def remove_component(self, key):
        """Remove component from registry
        """

        return self.kvstore.remove(key)

    def clear_components(self):
        """Clear all of the registered components
        """

        ComponentRegistry._component_overlays = {}

        for key in self.list_components():
            self.remove_component(key)

    def clear(self):
        """Clear all data from the registry
        """

        self.kvstore.clear()

    def list_components(self):
        """List all of the registered component names.

        This list will include all of the permanently stored components as
        well as any temporary components that were added with a temporary=True
        flag in this session.

        Returns:
            list of str: The list of component names.

            Any of these names can be passed to get_component as is to get the
            corresponding IOTile object.
        """

        overlays = list(self._component_overlays)
        items = self.kvstore.get_all()

        return overlays + [x[0] for x in items if not x[0].startswith('config:')]

    def iter_components(self):
        """Iterate over all defined components yielding IOTile objects."""

        names = self.list_components()

        for name in names:
            yield self.get_component(name)

    def list_config(self):
        """List all of the configuration variables
        """
        items = self.kvstore.get_all()
        return ["{0}={1}".format(x[0][len('config:'):], x[1])
                for x in items if x[0].startswith('config:')]

    def set_config(self, key, value):
        """Set a persistent config key to a value, stored in the registry

        Args:
            key (string): The key name
            value (string): The key value
        """

        keyname = "config:" + key

        self.kvstore.set(keyname, value)

    def get_config(self, key, default=MISSING):
        """Get the value of a persistent config key from the registry

        If no default is specified and the key is not found ArgumentError is raised.

        Args:
            key (string): The key name to fetch
            default (string): an optional value to be returned if key cannot be found

        Returns:
            string: the key's value
        """

        keyname = "config:" + key

        try:
            return self.kvstore.get(keyname)
        except KeyError:
            if default is MISSING:
                raise ArgumentError("No config value found for key", key=key)

            return default

    def clear_config(self, key):
        """Remove a persistent config key from the registry

        Args:
            key (string): The key name
        """

        keyname = "config:" + key
        self.kvstore.remove(keyname)


def _registry_folder(folder=None):
    if folder is None:
        folder = settings_directory()

        #If we are relative to a virtual environment, place the registry into that virtual env
        #Support both virtualenv and pythnon 3 venv
        if hasattr(sys, 'real_prefix'):
            folder = sys.prefix
        elif hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
            folder = sys.prefix

    return folder


def _check_registry_type(folder=None):
    """Check if the user has placed a registry_type.txt file to choose the registry type

    If a default registry type file is found, the DefaultBackingType and DefaultBackingFile
    class parameters in ComponentRegistry are updated accordingly.

    Args:
        folder (string): The folder that we should check for a default registry type
    """

    folder = _registry_folder(folder)

    default_file = os.path.join(folder, 'registry_type.txt')

    try:
        with open(default_file, "r") as infile:
            data = infile.read()
            data = data.strip()

            ComponentRegistry.SetBackingStore(data)
    except IOError:
        pass


def _ensure_package_loaded(path, component):
    """Ensure that the given module is loaded as a submodule.

    Returns:
        str: The name that the module should be imported as.
    """

    logger = logging.getLogger(__name__)

    packages = component.find_products('support_package')
    if len(packages) == 0:
        return None
    elif len(packages) > 1:
        raise ExternalError("Component had multiple products declared as 'support_package", products=packages)

    if len(path) > 2 and ':' in path[2:]:  # Don't flag windows C: type paths
        path, _, _ = path.rpartition(":")

    package_base = packages[0]
    relative_path = os.path.normpath(os.path.relpath(path, start=package_base))
    if relative_path.startswith('..'):
        raise ExternalError("Component had python product output of support_package", package=package_base, product=path)

    if not relative_path.endswith('.py'):
        raise ExternalError("Python product did not end with .py", path=path)

    relative_path = relative_path[:-3]
    if os.pathsep in relative_path:
        raise ExternalError("Python support wheels with multiple subpackages not yet supported", relative_path=relative_path)

    support_distro = component.support_distribution
    if support_distro not in sys.modules:
        logger.debug("Creating dynamic support wheel package: %s", support_distro)
        file, path, desc = imp.find_module(os.path.basename(package_base), [os.path.dirname(package_base)])
        imp.load_module(support_distro, file, path, desc)

    return "{}.{}".format(support_distro, relative_path)


def _try_load_module(path, import_name=None):
    """Try to programmatically load a python module by path.

    Path should point to a python file (optionally without the .py) at the
    end.  If it ends in a :<name> then name must point to an object defined in
    the module, which is returned instead of the module itself.

    Args:
        path (str): The path of the module to load
        import_name (str): The explicity name that the module should be given.
            If not specified, this defaults to being the basename() of
            path.  However, if the module is inside of a support package,
            you should pass the correct name so that relative imports
            proceed correctly.

    Returns:
        str, object: The basename of the module loaded and the requested object.
    """

    logger = logging.getLogger(__name__)

    obj_name = None
    if len(path) > 2 and ':' in path[2:]:  # Don't flag windows C: type paths
        path, _, obj_name = path.rpartition(":")

    folder, basename = os.path.split(path)
    if folder == '':
        folder = './'

    if basename == '' or not os.path.exists(path):
        raise ArgumentError("Could not find python module to load extension", path=path)

    basename, ext = os.path.splitext(basename)
    if ext not in (".py", ".pyc", ""):
        raise ArgumentError("Attempted to load module is not a python package or module (.py or .pyc)", path=path)

    if import_name is None:
        import_name = basename
    else:
        logger.debug("Importing module as subpackage: %s", import_name)

    try:
        fileobj = None
        fileobj, pathname, description = imp.find_module(basename, [folder])

        #Don't load modules twice
        if basename in sys.modules:
            mod = sys.modules[basename]
        else:
            mod = imp.load_module(import_name, fileobj, pathname, description)

        if obj_name is not None:
            if obj_name not in mod.__dict__:
                raise ArgumentError("Cannot find named object '%s' inside module '%s'" % (obj_name, basename), path=path)

            mod = mod.__dict__[obj_name]

        return (basename, mod)
    finally:
        if fileobj is not None:
            fileobj.close()


# Update our default registry backing store appropriately
_check_registry_type()



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

from .iotileobj import IOTile
from .registry import ComponentRegistry

__all__ = ['IOTile', 'ComponentRegistry']



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

# external_proxy.py
# Routines for importing proxy modules from registered components on your computer

from iotile.core.dev.registry import ComponentRegistry
from iotile.core.exceptions import DataError
from .plugin import TileBusProxyPlugin


def find_proxy_plugin(component, plugin_name):
    """ Attempt to find a proxy plugin provided by a specific component

    Args:
        component (string): The name of the component that provides the plugin
        plugin_name (string): The name of the plugin to load

    Returns:
        TileBuxProxyPlugin: The plugin, if found, otherwise raises DataError
    """

    reg = ComponentRegistry()

    plugins = reg.load_extensions('iotile.proxy_plugin', comp_filter=component, class_filter=TileBusProxyPlugin,
                                  product_name='proxy_plugin')

    for _name, plugin in plugins:
        if plugin.__name__ == plugin_name:
            return plugin

    raise DataError("Could not find proxy plugin module in registered components or installed distributions", component=component, name=plugin_name)
